#include "ns_class.h"

classattrib::classattrib()
{
  strcpy(dev, "");
  strcpy(name, "");
  bytes = 0;
  alive = 0;
  low = 0;
  ceil = 0;
  rate = 0;
  htb_rate = 0;
  htb_ceil = 0;
  classid = 0;
  htb_prio = 5;
  hold = 30;
  filter = u32;
  old_htb_ceil = 0;
  old_htb_rate = 0;
  work = false;
  use_ipt = true;
  use_tc_class = true;
  use_tc_filter = true;
  stats_show_htb_ceil = true;
  stats_show_rate = true;
  sum_in_section = true;
  class_interactive = false;
  local_section_id = section_id;
};

classattrib::~classattrib()
{
  for (unsigned int n = 0; n < tcfilter.size(); n++) {
    delete tcfilter[n];
  }
  tcfilter.clear();
}
int classattrib::store(char buf[])
{
  char option[35] = "";

  if (valid_whiteline(buf)) return 0;
  strcpy(option, awk(buf, 1));
  if (!strcmp(option, "class")) {
    strncpy(dev, awk(buf, 3), 6);
    strncpy(name, awk(buf, 4), 20);
    dev[MAX_IFACE - 1] = '\0';
    name[19] = '\0';
    if ( valid_dev(dev) == -1 ) {
      printf ( "\tNot proper dev: %s\n",buf );
      return -1;
    }
    if (strstr(markifaces, dev)) filter = fw;
  } 
  else if (!strcmp(option, "classid")) classid = atoi(awk(buf, 2));
  else if (!strcmp(option, "match")) tcfilter.push_back(new filterattrib(buf));
  else if (!strcmp(option, "filterid")) tcfilter[(tcfilter.size() - 1)]->store(buf);
  else if (!strcmp(option, "low")) low = unitconvert(awk(buf, 2), bit);
  else if (!strcmp(option, "ceil")) htb_ceil = ceil = unitconvert(awk(buf, 2), bit);
  else if (!strcmp(option, "rate")) { low = unitconvert(awk(buf, 2), bit); htb_ceil = ceil = unitconvert(awk(buf, 2), bit);} 
  else if (!strcmp(option, "interactive")) class_interactive = true;  
  else if (!strcmp(option, "htb-prio")) htb_prio = atoi(awk(buf, 2));
  else if (!strcmp(option, "htb-scheduler")) {
    if (!strcmp(awk(buf, 2), "none")) qdisc = none;
    else if (!strcmp(awk(buf, 2), "sfq")) qdisc = sfq;
  } 
  else if (!strcmp(option, "hold")) {
    if (!strlen(awk(buf, 2))) hold = 0;
    else hold = atoi(awk(buf, 2));
  }
  else if (!strcmp(awk(buf, 3), "show")) {
    if (!strcmp(awk(buf, 4), "stats")) {
      stats_show_htb_ceil = false;
      stats_show_rate = false;
    }
  } 
  else if (!strcmp(option, "do-not-shape")) {
    stats_show_htb_ceil = false;
    stats_show_rate = true;
    use_tc_class = false;    
    use_tc_filter = true;
    sum_in_section = false;
    low = 0;
	ceil = 0;
    htb_rate = 0;
    htb_ceil = 0;
  } 
  else if (!strcmp(option, "virtual")) {
    stats_show_htb_ceil = false;
    stats_show_rate = true;
    use_tc_class = false;    
    use_tc_filter = false;
    sum_in_section = false;
    low = 0;
	ceil = 0;
    htb_rate = 0;
    htb_ceil = 0;
  } 
  else if (!strcmp(option, "wrapper")) {
    stats_show_htb_ceil = true;
    stats_show_rate = true;
    use_tc_filter = true;
    use_tc_class = true;
    local_section_id = 1;
    sum_in_section = false;
	low = ceil;
  } 
  else {
    for (unsigned int i = 0; i < tcfilter.size(); i++) {
      if (tcfilter[i]->store(buf) == -1) { printf("%s: unrecognized option: %s\n", section, buf); return -1; }
    }
  }

  if (strcmp(option, "match")) {
    for (unsigned int i = 0; i < tcfilter.size(); i++) tcfilter[i]->store(buf);
  }    

  return 1;
}

int classattrib::prepare_tcclass(unsigned int classes)
{
  permill_bits = (float)(ceil-low)/1000;
  if ( use_tc_class ) {
    unsigned int htb_rate = section_shape / classes;
    if ( !sum_in_section ) htb_rate = htb_ceil;
    sprintf(tcclass_add, "tc class add dev %s parent 1:%d classid 1:%d htb rate %d ceil %d prio %d quantum %d", dev, local_section_id, classid, htb_rate, htb_rate, htb_prio, compute_quantum(htb_rate));
    sprintf(tcclass_del, "tc class del dev %s classid 1:%d", dev, classid);
    if ( qdisc == sfq ) {
      sprintf(tcqdisc_add, "tc qdisc add dev %s parent 1:%d sfq perturb 10", dev, classid);
      sprintf(tcqdisc_del, "tc qdisc del dev %s parent 1:%d", dev, classid);
    }
    else if ( qdisc == none ) {
      strcpy (tcqdisc_add , "" );
      strcpy (tcqdisc_del , "" );
    }
  }
  if ( use_tc_filter ) {
    for (unsigned int i = 0; i < tcfilter.size(); i++) {
      if (tcfilter[i]->prepare_tcfilter() == -1) return -1;
    }
  }
  return 1;
}

int classattrib::sysinit_iptables()
{
  for (unsigned int i = 0; i < tcfilter.size(); i++) {
    if (tcfilter[i]->sysinit_iptables() == -1) return -1;
  }
  return 0;
}

unsigned int classattrib::check(std::vector < char * > &iptables_summary, struct timeval now, float cycle_time)
{
  rate = 0;
  for (unsigned int i = 0; i < tcfilter.size(); i++) rate += tcfilter[i]->check(iptables_summary);
  if (work) {
    if (rate) alive = now.tv_sec;
    else if (((now.tv_sec - alive) >= hold) && (hold)) del();
  }
  if (!work) {
    if (rate) { add(); alive = now.tv_sec; }
  }
  rate = (unsigned int) ((float) rate / cycle_time);
  if ( sum_in_section) {
	if ( rate > htb_ceil ) rate = htb_ceil;
    return rate;
  }
  else return 0;
}

unsigned int classattrib::turn_back_usage_percent()
{
  if ( sum_in_section ) {
    if ( rate < htb_ceil ) return ( rate * 100 / htb_ceil );
    else return 100;
  } 
  else {
    return 0;
  }
}

unsigned int classattrib::turn_back_rate()
{
  if ( sum_in_section ) {
    if (rate < htb_ceil) return rate;
    else return htb_ceil;
  } 
  else {
    return 0;
  }
}

unsigned int classattrib::expect_h5n1_rate( unsigned int expect_rate )
{
  if ( sum_in_section ) { 
    if ( rate < expect_rate ) {
      unsigned int inc = 4;
      return ( rate + (power(100*rate/htb_ceil,inc) / power(100,inc-1)) * ((expect_rate-htb_ceil)/100) );
    }
    else {
      return expect_rate; 
    }
  }
  else return 0;
}

int classattrib::add()
{
  work = true;
  rate = htb_ceil;
  old_htb_ceil = htb_ceil;
  old_htb_rate = 0;
if (sum_in_section) working++;
  if (use_tc_class) {
    system_tc(tcclass_add);
    if (qdisc != none) system_tc(tcqdisc_add);
  }
  if (use_tc_filter) {  
    if (filter == u32) for (unsigned int i = 0; i < tcfilter.size(); i++) tcfilter[i]->add();
    if (filter == fw) tcfilter[0]->add();
  }
  return 1;
}

int classattrib::del()
{
  work = false;
  if (use_tc_filter) {  
    if (filter == u32) for (unsigned int i = 0; i < tcfilter.size(); i++) tcfilter[i]->del();
    if (filter == fw) tcfilter[0]->del();
  }
  if (sum_in_section) working--;
  if (use_tc_class) {
    if (qdisc != none) system_tc(tcqdisc_del);
    system_tc(tcclass_del);
  }
  return 1;
}

int classattrib::change()
{
  char buf[LONG_BUF];

  if ( !work ) return 0;
  if ( !sum_in_section ) return 0;
  if ( htb_ceil > ceil ) htb_ceil = ceil;
  if ( htb_ceil < low ) htb_ceil = low;

  unsigned int htb_rate = section_shape / working;  
  if ( htb_ceil < htb_rate ) htb_rate = htb_ceil;

  if (( old_htb_ceil == htb_ceil ) && ( old_htb_rate == htb_rate )) return 0;

  sprintf(buf, "tc class change dev %s classid 1:%d htb rate %d ceil %d prio %d quantum %d", dev, classid, htb_rate, htb_ceil, htb_prio, compute_quantum(htb_ceil));
  system_tc(buf);

  old_htb_rate = htb_rate;
  old_htb_ceil = htb_ceil;
  return 1;
}

void classattrib::stats(int pipe_write)
{
  char buf[LONG_BUF] = "";
  
  if (!work) return;
  
  if (stats_show_htb_ceil && stats_show_rate) sprintf(buf, "%s %d %d\n", name, htb_ceil, rate);
  else if (stats_show_htb_ceil) sprintf(buf, "%s %d -\n", name, htb_ceil);
  else if (stats_show_rate) sprintf(buf, "%s - %d\n", name, rate);
  write(pipe_write, buf, strlen(buf));
}
